using DifferentialEquations
using PyPlot
PyPlot.ioff()
using PyCall
cycler = pyimport("cycler")
using LaTeXStrings
using Printf
using CSV
using QuadGK
using LinearAlgebra
using Roots

Lpnames = [
L"k_{\rm bind}",
L"d_{\rm V}",    
L"k_{\rm diss}",
L"k_{\rm fuse}",
L"k_{\rm uncoat}",
L"d_{\rm endosome}",
L"d_{\rm gRNA}",
L"k_{\rm transl}f_{\rm ORF1}",
L"d_{\rm NSP}",
L"k_{\rm tr_{(-)}}",
L"K_{\rm NSP}",
L"d_{\rm gRNA_{(-)}}",
L"k_{\rm tr_{(+)}}",
L"k_{\rm complex}",
L"k_{\rm complex}n_{\rm N}",
L"k_{\rm transl}f_{\rm N}", 
L"k_{\rm transl}f_{\rm SP}",
L"d_{\rm N}",
L"d_{\rm SP}",
L"K_{\rm V_{rel}}n_{\rm SP}",
L"K_{\rm N}",
L"k_{\rm assemb}",
L"k_{\rm assemb}n_{\rm SP}",
L"d_{\rm N_{^{-}}gRNA}",
L"k_{\rm release}",
L"d_{\rm assembled}"
]

function SARS_CoV2_replication(du,u,p,t)
    V_free, V_bound, V_endosome, gRNA_plus, NSP,
    gRNA_minus, gRNA, N, SP, N_gRNA,
    V_assembled, V_released, V_cumulative = u

    (k_bind, d_V, k_diss, k_fuse, k_uncoat, d_endosome,
     d_gRNA, k_transl_f_ORF1, d_NSP, k_tr_minus, K_NSP,
     d_gRNA_minus, k_tr_plus, k_complex, k_complex_n_N,
     k_transl_f_N, k_transl_f_SP, d_N, d_SP,
     K_V_rel_n_SP, K_V_rel_n_N, k_assembl, k_assembl_n_SP,
     d_N_gRNA, k_release, d_assembled) = p

    #dV_free
    du[1] = -k_bind*V_free - d_V*V_free + k_diss*V_bound

    #dV_bound
    du[2] = k_bind*V_free - (k_fuse + k_diss + d_V)*V_bound

    #dV_endosome
    du[3] = k_fuse*V_bound - (k_uncoat + d_endosome)*V_endosome

    #dgRNA_plus
    du[4] = k_uncoat*V_endosome - d_gRNA*gRNA_plus

    #dNSP
    du[5] = k_transl_f_ORF1*gRNA_plus - d_NSP*NSP

    #dgRNA_minus
    du[6] = k_tr_minus*gRNA_plus*NSP/(K_NSP + NSP) - d_gRNA_minus*gRNA_minus

    f_assemb = SP/(SP + K_V_rel_n_SP)*N_gRNA
    f_complex = N*gRNA/(N + K_V_rel_n_N)

    #dgRNA
    du[7] = k_tr_plus*gRNA_minus*NSP/(K_NSP + NSP) - k_complex*f_complex - d_gRNA*gRNA

    #dN
    du[8] = k_transl_f_N*gRNA - k_complex_n_N*f_complex - d_N*N

    #dSP
    du[9] = k_transl_f_SP*gRNA - k_assembl_n_SP*f_assemb - d_SP*SP

    #dN_gRNA
    du[10] = k_complex*f_complex - k_assembl*f_assemb - d_N_gRNA*N_gRNA

    #dV_assembled
    du[11] = k_assembl*f_assemb - (k_release + d_assembled)*V_assembled

    #dV_released
    du[12] = k_release*V_assembled - d_V*V_released

    #cumulative virion release
    du[13] = V_released
end

# parameters
pt = CSV.File("parameters_lumped.txt", header=false, delim=' ', ignorerepeated=true)
pnames = String.(pt.Column1)
p = Float64.(pt.Column2)
lb = Float64.(pt.Column3)
ub = Float64.(pt.Column4)
@assert all(lb .<= p .<= ub)

println("Obtaining the numerical solution of the model")

# initial condition
u0 = repeat([0.0],13)
V0 = 10.0
u0[1] = V0

# solve the model
t_final = 24.0
tspan = (0.0, t_final)
prob = ODEProblem(SARS_CoV2_replication,u0,tspan,p)
sol = solve(prob,KenCarp5(),reltol=1e-16,abstol=1e-16)

# draw numerical solution
sL = ["V_{free}", "V_{bound}", "V_{endosome}", "gRNA_{(+)}",
     "NSP", "gRNA_{(-)}", "gRNA", "N", "SP", "N-gRNA",
     "V_{assembled}", "V_{released}"]

L = [latexstring("[{\\rm $s}]") for s in sL]
push!(L, L"\int_0^{t} [{\rm V_{released}}] dt")
L[10] = L"$[{\rm N}$-${\rm gRNA}]$"


function draw_sol(sol, filename)
    i2mm = 0.039370077777777776
    function mydraw(disp_vars, tspan; loc="upper left", ythresh=1.0, ythscale=0.4,
                                      lstyle="-", colors=nothing)
        if isnothing(colors)
            pl = ax.plot(sol.t, Array(sol')[:,disp_vars], linestyle=lstyle, linewidth=2.0)
        else
            for (var,col) in zip(disp_vars,colors)
                pl = ax.plot(sol.t, Array(sol')[:,var], linestyle=lstyle, color=col, linewidth=2.0)
            end
        end
        ax.set_xlim(tspan)
        ax.set_xlabel("t, hours")
        ax.set_ylim((0,1.15*ax.get_ylim()[2]))
        ax.set_yscale("symlog", linthresh=ythresh, linscale=ythscale, subs=2:9)
        ax.minorticks_on()
        ax.grid(which="major", linestyle="-", linewidth=0.1, color="lightgray")
        ax.tick_params(width=0.1, which="minor")
        ax.legend(loc=loc, labels=L[disp_vars])
        return pl
    end
    ##ax.yaxis.set_minor_locator(matplotlib.ticker.SymmetricalLogLocator(base=10.0,linthresh=ythresh))
    #ax.grid(which="minor", linestyle=":", linewidth=0.01)

    PyPlot.ioff()

    PyPlot.rc("text", usetex=false)
    PyPlot.rc("font", size=12)

    fig, axes = subplots(3,2,
         figsize=(210.0*i2mm,270.0*i2mm))
         #constrained_layout=true

    ax = axes[1,1]
    mydraw([1,2], [0,5], loc="upper right")
    ax.set_ylim((0,10))
    ax.set_yscale("linear")
    ax.set_ylabel("particles")
    ax.set_title("Binding and fusion")

    ax = axes[1,2]
    mydraw([3,4], [0,24], loc="upper right")
    ax.set_yscale("linear")
    ax.set_ylabel("numbers")
    ax.set_title("Endocytosis and uncoating")

    ax = axes[2,1]
    mydraw([5,6], [0,24], loc="upper left")
    ax.set_yscale("linear")
    ax.set_ylabel("molecules")
    ax.set_title(L"ORF1 translation and gRNA$_{(-)}$ synthesis")

    ax = axes[2,2]
    mydraw([7,8], [0,24], loc="lower right") #, ythresh=0.1
    ax.set_ylabel("molecules")
    ax.set_title("Transcription and translation")

    ax = axes[3,1]
    mydraw([9,10], [0,24], loc="lower right") #, ythresh=0.1
    ax.set_ylabel("molecules")
    ax.set_title("Translation and nucleocapsid formation")

    ax = axes[3,2]
    mydraw([11,12], [0,24], loc="upper left")
    #ax.set_yscale("linear")
    ax.set_ylabel("particles")
    ax.set_title("Assembly and release")

    plt.tight_layout()
    PyPlot.savefig(filename)
    plt.close(fig)
end
draw_sol(sol, "plot_model.pdf")
