using DifferentialEquations
using PyPlot
PyPlot.ioff()
using PyCall
cycler = pyimport("cycler")
using LaTeXStrings
using Printf
using CSV
using QuadGK
using LinearAlgebra
using Roots
using QuasiMonteCarlo


function SARS_CoV2_replication(du,u,p,t)
    V_free, V_bound, V_endosome, gRNA_plus, NSP,
    gRNA_minus, gRNA, N, SP, N_gRNA,
    V_assembled, V_released, V_cumulative = u

    (k_bind, d_V, k_diss, k_fuse, k_uncoat, d_endosome,
     d_gRNA, k_transl, f_ORF1, d_NSP, k_tr_minus, K_NSP,
     d_gRNA_minus, k_tr_plus, k_complex, K_N,
     f_N, f_SP, d_N, d_SP,
     n_SP, n_N, K_V_rel, k_assembl,
     d_N_gRNA, k_release, d_assembled) = p

    #dV_free
    du[1] = -k_bind*V_free - d_V*V_free + k_diss*V_bound

    #dV_bound
    du[2] = k_bind*V_free - (k_fuse + k_diss + d_V)*V_bound

    #dV_endosome
    du[3] = k_fuse*V_bound - (k_uncoat + d_endosome)*V_endosome

    #dgRNA_plus
    du[4] = k_uncoat*V_endosome - d_gRNA*gRNA_plus

    #dNSP
    du[5] = k_transl*f_ORF1*gRNA_plus - d_NSP*NSP

    #dgRNA_minus
    du[6] = k_tr_minus*gRNA_plus*NSP/(K_NSP + NSP) - d_gRNA_minus*gRNA_minus

    f_assemb = SP/(SP + K_V_rel*n_SP)*N_gRNA
    f_complex = N*gRNA/(N + K_N)

    #dgRNA
    du[7] = k_tr_plus*gRNA_minus*NSP/(K_NSP + NSP) - k_complex*f_complex - d_gRNA*gRNA

    #dN
    du[8] = k_transl*f_N*gRNA - k_complex*n_N*f_complex - d_N*N

    #dSP
    du[9] = k_transl*f_SP*gRNA - k_assembl*n_SP*f_assemb - d_SP*SP

    #dN_gRNA
    du[10] = k_complex*f_complex - k_assembl*f_assemb - d_N_gRNA*N_gRNA

    #dV_assembled
    du[11] = k_assembl*f_assemb - (k_release + d_assembled)*V_assembled

    #dV_released
    du[12] = k_release*V_assembled - d_V*V_released

    #cumulative virion release
    du[13] = V_released
end

# parameters
pt = CSV.File("parameters.txt", header=false, delim=' ', ignorerepeated=true)
pnames = String.(pt.Column1)
p = Float64.(pt.Column2)
lb = Float64.(pt.Column3)
ub = Float64.(pt.Column4)
@assert all(lb .<= p .<= ub)

println("Obtaining the numerical solution of the model")

# initial condition
u0 = repeat([0.0],13)
V0 = 10.0
u0[1] = V0

# solve the model
t_final = 24.0
tspan = (0.0, t_final)
prob = ODEProblem(SARS_CoV2_replication,u0,tspan,p)
sol = solve(prob,KenCarp5(),reltol=1e-16,abstol=1e-16)

println("Uncertainty analysis")

n = 1000
#n = 10000

s_parameters = QuasiMonteCarlo.sample(n,lb,ub,LatinHypercubeSample())

function prob_func(prob,i,repeat)
    remake(prob,p=s_parameters[:,i])
end

ensemble_prob = EnsembleProblem(prob, prob_func=prob_func)
sim = solve(ensemble_prob,KenCarp5(),reltol=1e-16,abstol=1e-16,EnsembleThreads(),trajectories=n)

summ = EnsembleSummary(sim,0:0.1:24)

tt = 0:0.01:24
ylow = DifferentialEquations.EnsembleAnalysis.timeseries_point_quantile(sim,0.05,tt)
ymed = DifferentialEquations.EnsembleAnalysis.timeseries_point_quantile(sim,0.5,tt)
yhigh = DifferentialEquations.EnsembleAnalysis.timeseries_point_quantile(sim,0.95,tt)

# progeny
Vlow = ylow[12,:]
Vmed = ymed[12,:]
Vhigh = yhigh[12,:]

PyPlot.ioff()

PyPlot.rc("text", usetex=false)
PyPlot.rc("font", size=13)

PyPlot.figure(figsize=(6.4,5.4), constrained_layout=true)

PyPlot.plot(tt, Vmed, label=L"median ($[{\rm V_{released}}]({\rm t})$)")
PyPlot.fill_between(tt, Vlow, Vhigh, alpha=0.2, label="5%-95% confidence region")
PyPlot.yscale("symlog", subs=2:9, linthresh=1.0, linscale=0.25)
PyPlot.minorticks_on()
PyPlot.grid(which="major", linestyle="-", linewidth=0.1, color="lightgray")
PyPlot.legend(loc="upper left")
PyPlot.xlim((0,24))
PyPlot.xlabel("t, hours")
PyPlot.ylabel("virions")
PyPlot.title("Parameter uncertainty in progeny release kinetics")

#PyPlot.tight_layout()
#PyPlot.show()
PyPlot.savefig("uncertainty_progeny.pdf")
PyPlot.close()


n = 11
#s_Vfree = QuasiMonteCarlo.sample(n,5.0,15.0,GridSample(1.0))
s_Vfree = 5:1:15
function prob_func2(prob,i,repeat)
    u0_ = similar(prob.u0)
    u0_[:] .= 0.0
    u0_[1] = s_Vfree[i]
    remake(prob,u0=u0_)
end

ensemble_prob2 = EnsembleProblem(prob, prob_func=prob_func2)
sim2 = solve(ensemble_prob2,KenCarp5(),reltol=1e-16,abstol=1e-16,EnsembleThreads(),trajectories=n)

tt = 0:0.01:24

PyPlot.ioff()

PyPlot.rc("text", usetex=false)
PyPlot.rc("font", size=13)

PyPlot.figure(figsize=(6.4,5.4), constrained_layout=true)

for i in 1:n
    PyPlot.plot(tt, sim2[i](tt,idxs=12), label=latexstring("\$[{\\rm V_{free}}](0) = \$ $(Int(sim2[i].prob.u0[1]))"))
end

PyPlot.yscale("symlog", subs=2:9, linthresh=1.0, linscale=0.2)
PyPlot.minorticks_on()
PyPlot.grid(which="major", linestyle="-", linewidth=0.1, color="lightgray")
PyPlot.legend(loc="upper left")
PyPlot.xlim((0,24))
PyPlot.xlabel("t, hours")
PyPlot.ylabel("virions")
PyPlot.title(L"Progeny release kinetics at various $[{\rm V_{free}}](0)$")

#PyPlot.tight_layout()
#PyPlot.show()
PyPlot.savefig("uncertainty_progeny_initconditions.pdf")
PyPlot.close()
